#include <sys/types.h>
#include <unistd.h>
#include <getopt.h>
#include <time.h>

#include "version.h"
#include "process.h"
#include "rrdtool.h"
#include "parser.h"
#include "debug.h"

#include "config.h"
#include "built_in_parsers.h"

#include "plugin_manager.h"

struct user_options {
	int max_jobs;
	char *config_file;
};

int read_args(int argc, char *argv[], struct user_options *opts)
{
	int option_index = 0, c;
	static struct option long_options[] = {
		{ .val = 'j', .has_arg = 1, .name = "jobs", },
		{ .val = 'c', .has_arg = 1, .name = "config", },
		{ .val = 'l', .has_arg = 1, .name = "log-file", },
		{ .val = 'v', .name = "verbose", },
		{ .val = 'q', .name = "quiet", },
		{ },
	};
	char short_options[] = "j:c:l:vq";

	while (1) {
		c = getopt_long(argc, argv, short_options, long_options,
				&option_index);

		if (c == -1)
			break;

		switch (c) {
		case 'j':
			opts->max_jobs = atoi(optarg);
			break;

		case 'c':
			opts->config_file = optarg;
			break;

		case 'l':
			open_log_file(optarg);
			break;

		case 'v':
			trace_level++;
			break;

		case 'q':
			trace_level--;
			break;

		case '?':
			return -1;
		}
	}

	while (optind < argc) {
		opts->config_file = argv[optind];
		optind++;
	}

	return 0;
}

int main(int argc, char *argv[])
{
	struct user_options opts;
	struct rrd_database *db, **db_list = NULL;
	int sleeptime;
	int ret = 0;

	pr_info("%s Version %s starting\n", argv[0], RRDD_VERSION);

	bzero(&opts, sizeof(opts));

	if (read_args(argc, argv, &opts) < 0)
		return -1;

	init_plugin_manager(argv[0]);
	register_built_in_parsers();

	if (!opts.config_file) {
		pr_err("No database config file given. Nothing to do\n");
		return 1;
	}

	db_list = populate_database(opts.config_file);

	if (ret || !db_list)
		return 1;

	if (rrdtool_create_missing_databases(db_list))
		return 1;

	if (init_jobcontrol(opts.max_jobs))
		return -1;

	while (1) {
		const char *db_name;
		char timestr[128];
		time_t t;

		/*
		 * Update all databases parallel in one shot
		 */
		while ((db = get_outdated_db(db_list)))
			rrdtool_update_data(db);

		sleeptime = get_next_update(db_list, &db_name);

		t = time(0) + sleeptime;
		strftime(timestr, sizeof(timestr), "%T", localtime(&t));
		pr_info("Next scheduled event \"%s\" at %s, in %d seconds\n",
			db_name, timestr, sleeptime);

		poll_job_requests(sleeptime);
	}
	return 0;
}
